/*****************************************************************************
    fileio.c

    Diese Datei ist Teil des Programmpakets 'ACL-tftp'.
    Alle Rechte bei Andreas G. Schindler
                    Klappacher Str. 2i
                    64285 Darmstadt, Deutschland.

    Dieses Programm ist 'Freie Software'; Sie knnen es entsprechend den 
    Bedingungen der 'Allgemeinen ffentlichen GNU-Lizenz' (verffentlicht 
    durch die 'Free Software Foundation') weiter verbreiten und/oder 
    modifizieren. Massgeblich ist die Version 2 oder eine neuere Version
    dieser Lizenz. 

    Dieses Programm wird in der Hoffnung dass es ntzlich is weitergegeben,
    jedoch OHNE JEDE GARANTIE. Dieser Garantieauschluss bezieht sich auch
    auf Eigenschaften wie Marktfhigkeit oder Eignung zu einem bestimmten
    Zweck. Details finden sich in der 'Allgemeinen ffentlichen GNU-Lizenz'.

    Eine Kopie dieser Lizenz sollte in  Programmpaket, dem diese Datei
    entstammt, enthalten sein. Falls nicht, schreiben Sie an die 
    'Free Software Foundation, Inc.', 59 Temple Place, Suite 330, Boston, 
    MA  02111-1307  USA.


    This file is part of the 'ACL-tftp' Package
    Copyright (C) 1999-2003 by Andreas G. Schindler
                               Klappacher Str. 2i
                               Darmstadt, Germany.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************************************************/
#include "auto.h"
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

#include "params.h"
#include "netio.h"
#include "fileio.h"
#include "globals.h"
#include "logfun.h"

#ifndef IOBUFMIN
#define IOBUFMIN (64*1024)	/* minimum read/write ahead buffer size */
#endif

#ifndef IOBUFMAX
#define IOBUFMAX (2*1024*1024)	/* maximum read/write ahead buffer size */
#endif

static char *rahBuffer = NULL;
static int  intBUFSIZ  = 0;
static char *next;
static int  avail, lfdone, ignorecr, outlen;


u_long currentSize(int fd)
{
    return (u_long)lseek(fd, 0, SEEK_CUR);
}


int initXferBuffer(int filesize)
{
    int newsize;

    newsize = (filesize < IOBUFMAX) ? filesize : IOBUFMAX;
    newsize = (newsize  > IOBUFMIN) ? newsize  : IOBUFMIN;

    if (newsize > intBUFSIZ) {
        intBUFSIZ = newsize;
        if (rahBuffer != NULL) 
            rmXferBuffer();
    }
    if (rahBuffer == NULL) {
        if ((rahBuffer = (char *)Malloc(intBUFSIZ)) == NULL) {
            errNoDynmem;
            return 1;
        }
        LOG_MSG6 (LOG_DEBUG, "%s %d bytes allocated",__FUNCTION__,intBUFSIZ);
    }
    next  = rahBuffer;
    avail = lfdone = ignorecr = outlen = 0;
    return 0;
}


void rmXferBuffer (void)
{
    if (rahBuffer != NULL) {
        Free(rahBuffer);
        rahBuffer = NULL;
        LOG_MSG6 (LOG_DEBUG, "%s %d bytes released",__FUNCTION__,intBUFSIZ);
    }
}


int getXferBuffer (int fd, char *buf, int reqlen, int mode, count_t *iocount)
{
    int actlen = 0;

    while (reqlen > 0) {

        if (avail > 0) {	/* still something buffered ? */
            int siz;
            siz = (avail < reqlen) ? avail : reqlen;
            if (mode == RM_NETASCII) {	/* convert <lf> to <cr><lf> */
		while (siz--) {
                    if (*next == '\n' && !lfdone) {
                        *buf++  = '\r';
                        lfdone  = 1;
                    }
                    else {
                        *buf++  = *next++;
                        lfdone  = 0;
                        avail--;
                    }
                    actlen++;
                    reqlen--;
                }
            }
            else {			/* copy largest possible chunk */
                memcpy(buf, next, siz);
                actlen += siz;
                reqlen -= siz;
                avail  -= siz;
                next   += siz;
                buf    += siz;
            }
        }
        else {				/* get next chunk from disk */
            avail = read (fd, rahBuffer, intBUFSIZ);

            if (avail == 0)
                break;			/* end-of-file: return partial buffer */
            else if (avail < 0) {
                if (errno == EINTR) {
                    if (isAborted)	/* signal */
                        return -1;
                    else
                        avail = 0;
                }
                else
                    return -1;		/* read error: just quit */
            }
            next = rahBuffer;
            LOG_MSG6 (LOG_DEBUG, "%s %d Bytes read", __FUNCTION__, avail);
        }
    }
    *iocount += (count_t)actlen;
    LOG_MSG6 (LOG_DEBUG, "%s %d Bytes returned", __FUNCTION__, actlen);
    return actlen;
}


int putXferBuffer (int fd, char *buf, int reqlen, int mode, int flush,
                   count_t *iocount)
{
    int actlen = reqlen;

    LOG_MSG6 (LOG_DEBUG, "%s %d Bytes requested", __FUNCTION__, reqlen);

    while (reqlen > 0) {
        if (avail > 0) {	/* still someting buffered ? */

            int siz = (reqlen > avail) ? avail : reqlen;

            if (mode == RM_NETASCII) {	/* convert <cr><lf> to <lf> */

		while (siz--) {
                    if (*buf == '\r') {
                        ignorecr = 1;
                        buf++;
                        reqlen--;
                        continue;
                    }
                    if (ignorecr) {
                        ignorecr = 0;
                        if (*buf != '\n') {
                            *next++  = '\r';
                            avail--;
                            outlen++;
                            break;
                        }
                    }
                    *next++  = *buf++;
                    avail--;
                    outlen++;
                    reqlen--;
                }
            }
            else {		/* copy largest possible chunk */

                memcpy(next, buf, siz);
                outlen += siz;
                reqlen -= siz;
                avail  -= siz;
                next   += siz;
                buf    += siz;
            }
        }
        else {			/* write next chunk to disk */
            if (outlen)
                if (write(fd, rahBuffer, outlen) != outlen)
                    return -1;		/* write error: just quit */

            LOG_MSG6 (LOG_DEBUG, "%s %d Bytes written",__FUNCTION__,outlen);
            next   = rahBuffer;
            avail  = intBUFSIZ;
            outlen = 0;
        }
    }

    if (flush) {
        if (outlen) {
            if (write (fd, rahBuffer, outlen) != outlen)
                return -1;	/* write error: just quit */

            LOG_MSG6 (LOG_DEBUG, "%s %d Bytes written",__FUNCTION__,outlen);
        }
        next   = rahBuffer;
        avail  = intBUFSIZ;
        outlen = 0;
    }

    *iocount += (count_t)actlen;
    return actlen;
}


int file2buffer (char *file, char **buffer, ulong *bufsiz, int mode)
{
    int fd;
    int rtc = 0;
    int reqlen, actlen; 
    count_t iocount;
    char *bufptr;
    static char temp[8192];

    if ((fd = open(file, O_RDONLY)) < 0)
        return 1;

    if(initXferBuffer(maxCacheobj))
        return 1;

    bufptr = *buffer;
    reqlen = *bufsiz;
    actlen = 0;
/*
	at first getXferBuffer() try to fill the whole buffer,
	which should be preallocated to the size of the file.
	thus, in binary mode, only one single action is necessary.
	if some date is left (usually in netascii mode, because
	of the conversion from \n to \r\n), an auxiliary buffer is
	used to chatch the rest (maybe in several requests). The
	main buffer is then expanded to fit the additional data.
*/
    LOG_MSG6 (LOG_DEBUG, "%s Loading file '%s' in %s mode",
              __FUNCTION__, file, (mode == 2) ? "netascii" : "binary");

    for (;;) {
        rtc = getXferBuffer(fd, bufptr, reqlen, mode, &iocount);
        if (rtc < 0) {
            close (fd);
            rmXferBuffer();
            return 1;
        }
        actlen += rtc;
        if ((bufptr == temp) && rtc) {	/* additional data */
            *buffer = (char *)Realloc(*buffer, *bufsiz + rtc);
            if (*buffer == NULL) {
                errNoDynmem;
                close (fd);
                rmXferBuffer();
                return 1;
            }
            memcpy(*buffer + *bufsiz, temp, rtc);
            *bufsiz += rtc;
        }
        if (rtc < reqlen)	/* end of file	*/
            break;
        bufptr = temp;		/* be sure to use the aux buffer	*/
        reqlen = sizeof(temp);	/* at additional requests		*/
    }
    close (fd);
    *bufsiz = actlen;
    return 0;
}


#ifdef TEST

#include "globals.h"
#include "logfun.h"
#define TESTFILE "io_test.dat"

static FILE *testfd;

void mkTestfile(void)
{
    remove (TESTFILE);
    if ((testfd = fopen(TESTFILE, "w+")) == NULL) {
        printf("Cannot open test file -- exiting\n");
        exit(1);
    }

}

void removeTestfile (void)
{
    fclose(testfd);
    remove(TESTFILE);
}

static void filbuf (char *buf, int size, int mark)
{
    int i, j;

    if (size <= 4) {
        snprintf(buf, 4, "%3d",mark % 100);
        buf[3] = '\n';
    }
    else {
        snprintf(buf, 5, "%3d ",mark % 100);
        for (i = 4; i < 62; i++)
            buf[i] = '-';
        buf[62] = '\r';
        buf[63] = '\n';

        if (size > 64) {
            for (i = 64; i < size; i++)
            if ((i % 64) == 63)
                buf[i] = '\n';
            else if ((i % 64) == 62)
                buf[i] = '\r';
            else if ((i % 64) < 3)
                buf[i] = ' ';
            else {
                j = (i % 16);
                buf[i] = (j > 9) ? (j - 10 + 'a') : j + '0';
            }
        }
    }
}


static int docompare (int seg, char *buf1, char *buf2, int len)
{
    int i, s, e ;

    for (i = 0; i < len; i++) {
        if(buf1[i] != buf2[i]) {
            printf ("Data mismatch in segment %d (%d) offset %d: file:[0x%2.2x] comp:[0x%2.2x]\n",
                    seg, len, i, (int)buf1[i], (int)buf2[i]);
            s = (i-5) < 0   ? 0   : (i-5);
            e = (i+5) > len ? len : (i+5);
            printf ("[");
            for (i = s; i < e; i++)
                printf ("%c", buf1[i]);
            printf ("]\n[");
            for (i = s; i < e; i++)
                printf ("%c", buf2[i]);
            printf ("]\n");
            return 1;
        }
    }
    return 0;
}


int stripbuf (char *buf, int len)
{
    char *ptr;
    int count;

    for (ptr = buf, count = 0; len-- > 0; buf++) {
        if (*buf == '\r' && *(buf+1) == '\n')
            continue;
        *ptr++ = *buf;
        count++;
    } 
    return count;
}


int main (int argc, char *argv[])
{
#define NS 100

    int step = 0, s, ns;
    char segbuf[2*SEGSIZE], cmpbuf[2*SEGSIZE];
    int lens[NS], err = 0, rtc; 
    
    msgLevel  = 5;
    maxSyslog = -1;
    startLog(-1);

    printf ("Step %d: Writing a binary file of %d full segments\n", 
             ++step, ns = NS);
    mkTestfile();
    for (s = 0; s < ns; s++) {
        lens[s] = SEGSIZE;
        filbuf(segbuf, lens[s], s);
        rtc = putXferBuffer (fileno(testfd), segbuf, lens[s], RM_OCTET, (s == (ns-1)) ? 1 : 0, NULL);
        if (rtc != SEGSIZE) {
            printf ("Error writing segment %d\n", s);
            err = 1;
            break;
        }
    }
    rewind (testfd);
    if (!err) {
        for (s = 0; s < ns; s++) {
            filbuf(cmpbuf, lens[s], s);
            if (fread(segbuf, 1, lens[s], testfd) != lens[s]) {
                printf ("Error reading segment %d\n", s);
                err = 1;
                break;
            }
            if ((err = docompare (s, segbuf, cmpbuf, lens[s])))
                break;
        }
    }
    removeTestfile();

    printf ("Step %d: Writing a binary file of 1 half and %d full segments\n",
             ++step, (ns = NS) -1 );
    mkTestfile();
    for (s = 0; s < ns; s++) {
        lens[s] = (s) ? SEGSIZE : SEGSIZE/2;
        filbuf(segbuf, lens[s], s);
        rtc = putXferBuffer (fileno(testfd), segbuf, lens[s], RM_OCTET, (s == (ns-1)) ? 1 : 0, NULL);
        if (rtc != lens[s]) {
            printf ("Error writing segment %d\n", s);
            err = 1;
            break;
        }
    }
    rewind (testfd);
    if (!err) {
        for (s = 0; s < ns; s++) {
            filbuf(cmpbuf, lens[s], s);
            if (fread(segbuf, 1, lens[s], testfd) != lens[s]) {
                printf ("Error reading segment %d\n", s);
                err = 1;
                break;
            }
            if ((err = docompare (s, segbuf, cmpbuf, lens[s])))
                break;
        }
    }
    removeTestfile();

    printf ("Step %d: Reading a binary file of %d full segments\n", 
             ++step, ns = NS);
    mkTestfile();
    for (s = 0; s < ns; s++) {
        lens[s] = SEGSIZE;
        filbuf(segbuf, lens[s], s);
        if (fwrite(segbuf, 1, lens[s], testfd) != lens[s]) {
            printf ("Error writing segment %d\n", s);
            err = 1;
            break;
        }
    }
    rewind (testfd);
    if (!err) {
        for (s = 0; s < ns; s++) {
            filbuf(cmpbuf, lens[s], s);
            rtc = getXferBuffer (fileno(testfd), segbuf, lens[s], RM_OCTET, NULL);
            if (rtc != lens[s]) {
                printf ("Error reading segment %d\n", s);
                err = 1;
                break;
            }
            if ((err = docompare (s, segbuf, cmpbuf, lens[s])))
                break;
        }
    }
    if(getXferBuffer (fileno(testfd), segbuf, lens[s], RM_OCTET, NULL))
        printf ("No EOF -- dangling test file data");
    removeTestfile();

    printf ("Step %d: Reading a binary file of %d + 1/2 segments\n", 
             ++step, (ns = NS) -1);
    mkTestfile();
    for (s = 0; s < ns; s++) {
        lens[s] = (ns == (NS-1)) ? SEGSIZE/2 : SEGSIZE;
        filbuf(segbuf, lens[s], s);
        if (fwrite(segbuf, 1, lens[s], testfd) != lens[s]) {
            printf ("Error writing segment %d\n", s);
            err = 1;
            break;
        }
    }
    rewind (testfd);
    if (!err) {
        for (s = 0; s < ns; s++) {
            filbuf(cmpbuf, lens[s], s);
            rtc = getXferBuffer (fileno(testfd), segbuf, SEGSIZE, RM_OCTET, NULL);
            if (rtc != lens[s]) {
                printf ("Error reading segment %d\n", s);
                err = 1;
                break;
            }
            if ((err = docompare (s, segbuf, cmpbuf, lens[s])))
                break;
        }
    }
    if(getXferBuffer (fileno(testfd), segbuf, lens[s], RM_OCTET, NULL))
        printf ("No EOF -- dangling test file data");
    removeTestfile();

    printf ("Step %d: Writing a file with netascii conversion of %d full segments\n", 
             ++step, ns = NS);
    mkTestfile();
    for (s = 0; s < ns; s++) {
        lens[s] = SEGSIZE;
        filbuf(segbuf, lens[s], s);
        rtc = putXferBuffer (fileno(testfd), segbuf, lens[s], RM_NETASCII, (s == (ns-1)) ? 1 : 0, NULL);
        if (rtc != SEGSIZE) {
            printf ("Error writing segment %d\n", s);
            err = 1;
            break;
        }
    }
    rewind (testfd);
    if (!err) {
        for (s = 0; s < ns; s++) {
            filbuf(cmpbuf, lens[s], s);
            lens[s] = stripbuf(cmpbuf, lens[s]);
            if (fread(segbuf, 1, lens[s], testfd) != lens[s]) {
                printf ("Error reading segment %d\n", s);
                err = 1;
                break;
            }
            if ((err = docompare (s, segbuf, cmpbuf, lens[s])))
                break;
        }
    }
    removeTestfile();

    printf ("Step %d: Reading a file with netascii conversion of %d full segments\n", 
             ++step, ns = NS);
    mkTestfile();
    for (s = 0; s < ns; s++) {
        lens[s] = SEGSIZE;
        filbuf(segbuf, lens[s], s);
        lens[s] = stripbuf(segbuf, lens[s]);
        if (fwrite(segbuf, 1, lens[s], testfd) != lens[s]) {
            printf ("Error writing segment %d\n", s);
            err = 1;
            break;
        }
    }
    rewind (testfd);
    if (!err) {
        for (s = 0; s < ns; s++) {
            lens[s] = SEGSIZE;
            filbuf(cmpbuf, lens[s], s);
            rtc = getXferBuffer (fileno(testfd), segbuf, lens[s], RM_NETASCII, NULL);
            if (rtc != lens[s]) {
                printf ("Error reading segment %d\n", s);
                err = 1;
                break;
            }
            if ((err = docompare (s, segbuf, cmpbuf, lens[s])))
                break;
        }
    }
    if(getXferBuffer (fileno(testfd), segbuf, lens[s], RM_OCTET, NULL))
        printf ("No EOF -- dangling test file data");
    removeTestfile();


    return 0;
}

#endif /* TEST */
